# -*- coding: utf-8 -*-

import logging
import platform
import urllib
import ctypes
import os
import platform
import sys

import PySide2
from PySide2 import QtCore
from PySide2.QtGui import *
from PySide2.QtCore import *
from PySide2.QtWidgets import *

ld_library_path = os.environ.get("LD_LIBRARY_PATH")
from cefpython3 import cefpython as cef


def group(url, icon, title, size):
    sys.excepthook = ExceptHook
    settings = {}
    if WINDOWS:
        settings["external_message_pump"] = True
    elif MAC:
        # Issue #442 requires enabling message pump on Mac
        # in Qt example. Calling cef.DoMessageLoopWork in a timer
        # doesn't work anymore.
        settings["external_message_pump"] = True
    cef.Initialize(settings=settings,
                   commandLineSwitches={
                       "autoplay-policy": "no-user-gesture-required",
                       "lang": 'zh-CN'
                   })
    app = CefApp(url, icon, title, size)
    return app, app.frame.browser


def ExceptHook(exc_type, exc_value, exc_trace):
    '''
    原本的cef.ExceptHook打印不了中文，我拿來魔改一下。
    '''
    import traceback
    if exc_type == SystemExit:
        hint = '[CEF Python] SystemExit.'
        print('\n' + '='*len(hint) + '\n' + hint)
    else:
        hint = '[CEF Python] ExceptHook: catched exception, will shutdown CEF.'
        print('\n' + '='*len(hint) + '\n' + hint)
        msg = ''.join(traceback.format_exception(exc_type, exc_value, exc_trace))
        print('\n' + msg)
    cef.QuitMessageLoop()
    cef.Shutdown()
    os._exit(1)


# Fix for PyCharm hints warnings when using static methods
WindowUtils = cef.WindowUtils()

# Platforms
WINDOWS = (platform.system() == "Windows")
LINUX = (platform.system() == "Linux")
MAC = (platform.system() == "Darwin")


def main():
    check_versions()
    sys.excepthook = cef.ExceptHook  # To shutdown all CEF processes on error
    settings = {}
    if WINDOWS:
        settings["external_message_pump"] = True
    elif MAC:
        # Issue #442 requires enabling message pump on Mac
        # in Qt example. Calling cef.DoMessageLoopWork in a timer
        # doesn't work anymore.
        settings["external_message_pump"] = True

    cef.Initialize(settings)
    app = CefApp("http://localhost:8080", None, "localhost", (800, 600))
    app.MainLoop()
    del app.main_window  # Just to be safe, similarly to "del app"
    del app  # Must destroy app object before calling Shutdown
    cef.Shutdown()


def check_versions():
    print("[qt.py] CEF Python {ver}".format(ver=cef.__version__))
    print("[qt.py] Python {ver} {arch}".format(
        ver=platform.python_version(), arch=platform.architecture()[0]))
    print("[qt.py] PySide2 {v1} (qt {v2})".format(v1=PySide2.__version__,
                                                  v2=QtCore.__version__))
    # CEF Python version requirement
    assert cef.__version__ >= "55.4", "CEF Python v55.4+ required to run this"


class MainWindow(QMainWindow):
    def __init__(self, url, icon, title, size):
        # noinspection PyArgumentList
        super(MainWindow, self).__init__(None)
        self.url = url

        logging.debug("[qt.py] MainWindow DPI scaled size: %s" % str(size))
        width, height = tuple(size)
        self.resize(width, height)

        self.setWindowTitle(title)

        ic = QIcon(icon)
        self.setWindowIcon(ic)

        self.cef_widget = None
        self.setFocusPolicy(Qt.StrongFocus)
        self.setupLayout()

    def setupLayout(self):
        self.cef_widget = CefWidget(self)
        layout = QGridLayout()
        # noinspection PyArgumentList
        layout.addWidget(self.cef_widget, 1, 0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.setRowStretch(0, 0)
        layout.setRowStretch(1, 1)
        # noinspection PyArgumentList
        frame = QFrame()
        frame.setLayout(layout)
        self.setCentralWidget(frame)

        if WINDOWS:
            # On Windows with PyQt5 main window must be shown first
            # before CEF browser is embedded, otherwise window is
            # not resized and application hangs during resize.
            self.show()

        # Browser can be embedded only after layout was set up
        self.cef_widget.embed_browser(self.url)

        if LINUX:
            # On Linux with PyQt5 the QX11EmbedContainer widget is
            # no more available. An equivalent in Qt5 is to create
            # a hidden window, embed CEF browser in it and then
            # create a container for that hidden window and replace
            # cef widget in the layout with the container.
            # noinspection PyUnresolvedReferences, PyArgumentList
            self.container = QWidget.createWindowContainer(
                self.cef_widget.hidden_window, parent=self)
            # noinspection PyArgumentList
            layout.addWidget(self.container, 1, 0)

    def closeEvent(self, event):
        # Close browser (force=True) and free CEF reference
        if self.cef_widget.browser:
            self.cef_widget.browser.CloseBrowser(True)
            self.clear_browser_references()

    def clear_browser_references(self):
        # Clear browser references that you keep anywhere in your
        # code. All references must be cleared for CEF to shutdown cleanly.
        self.cef_widget.browser = None

    def set_browser_object(self, name, obj):
        bindings = cef.JavascriptBindings()
        bindings.SetObject(name, obj)
        self.cef_widget.browser.SetJavascriptBindings(bindings)

    def toggleFullScreen(self):
        if self.isFullScreen():
            self.showNormal()
        else:
            self.showFullScreen()

    @property
    def browser(self):
        return self.cef_widget.browser


class CefWidget(QWidget):
    def __init__(self, parent=None):
        # noinspection PyArgumentList
        super(CefWidget, self).__init__(parent)
        self.parent = parent
        self.browser = None
        self.hidden_window = None  # Required for PyQt5 on Linux
        self.show()

    def focusInEvent(self, event):
        # This event seems to never get called on Linux, as CEF is
        # stealing all focus due to Issue #284.
        if cef.GetAppSetting("debug"):
            print("[qt.py] CefWidget.focusInEvent")
        if self.browser:
            if WINDOWS:
                WindowUtils.OnSetFocus(self.getHandle(), 0, 0, 0)
            self.browser.SetFocus(True)

    def focusOutEvent(self, event):
        # This event seems to never get called on Linux, as CEF is
        # stealing all focus due to Issue #284.
        if cef.GetAppSetting("debug"):
            print("[qt.py] CefWidget.focusOutEvent")
        if self.browser:
            self.browser.SetFocus(False)

    def embed_browser(self, url):
        if LINUX:
            # noinspection PyUnresolvedReferences
            self.hidden_window = QWindow()
        window_info = cef.WindowInfo()
        rect = [0, 0, self.width(), self.height()]
        window_info.SetAsChild(self.getHandle(), rect)
        self.browser = cef.CreateBrowserSync(window_info, url=url, browserSettings={'web_security_disabled': True})
        self.browser.SetClientHandler(FocusHandler(self))

    def getHandle(self):
        if self.hidden_window:
            # PyQt5 on Linux
            return int(self.hidden_window.winId())
        try:
            # PyQt4 and PyQt5
            return int(self.winId())
        except:
            # PySide:
            # | QWidget.winId() returns <PyCObject object at 0x02FD8788>
            # | Converting it to int using ctypes.
            ctypes.pythonapi.PyCapsule_GetPointer.restype = (ctypes.c_void_p)
            ctypes.pythonapi.PyCapsule_GetPointer.argtypes = ([
                ctypes.py_object
            ])
            return ctypes.pythonapi.PyCapsule_GetPointer(self.winId(), None)

    def moveEvent(self, _):
        self.x = 0
        self.y = 0
        if self.browser:
            if WINDOWS:
                WindowUtils.OnSize(self.getHandle(), 0, 0, 0)
            elif LINUX:
                self.browser.SetBounds(self.x, self.y, self.width(),
                                       self.height())
            self.browser.NotifyMoveOrResizeStarted()

    def resizeEvent(self, event):
        size = event.size()
        if self.browser:
            if WINDOWS:
                WindowUtils.OnSize(self.getHandle(), 0, 0, 0)
            elif LINUX:
                self.browser.SetBounds(self.x, self.y, size.width(),
                                       size.height())
            self.browser.NotifyMoveOrResizeStarted()


class CefApp(QApplication):
    def __init__(self, url, icon, title, size):
        super(CefApp, self).__init__([])
        if cef.GetAppSetting("external_message_pump") or \
                cef.GetAppSetting("multi_threaded_message_loop"):
            self.timer = None
        else:
            self.timer = self.createTimer()

        self.url, self.icon, self.title, self.size = url, icon, title, size

        self.main_window = MainWindow(url, icon, title, size)
        self.main_window.show()
        self.main_window.activateWindow()
        self.main_window.raise_()
        self.frame = self.main_window

    def MainLoop(self):
        self.exec_()
        if self.timer is not None:
            self.stopTimer()

    def createTimer(self):
        timer = QTimer()
        # noinspection PyUnresolvedReferences
        timer.timeout.connect(self.onTimer)
        timer.start(10)
        return timer

    def onTimer(self):
        cef.MessageLoopWork()

    def stopTimer(self):
        # Stop the timer after Qt's message loop has ended
        self.timer.stop()


class FocusHandler(object):
    def __init__(self, cef_widget):
        self.cef_widget = cef_widget

    def OnTakeFocus(self, **_):
        if cef.GetAppSetting("debug"):
            print("[qt.py] FocusHandler.OnTakeFocus")

    def OnSetFocus(self, **_):
        if cef.GetAppSetting("debug"):
            print("[qt.py] FocusHandler.OnSetFocus")

    def OnGotFocus(self, browser, **_):
        if cef.GetAppSetting("debug"):
            print("[qt.py] FocusHandler.OnGotFocus")
        self.cef_widget.setFocus()
        # Temporary fix no. 1 for focus issues on Linux (Issue #284)
        if LINUX:
            browser.SetFocus(True)


if __name__ == '__main__':
    main()



import logging

import chardet


def good_open(path):
    with open(path, 'rb') as f:
        a = chardet.detect(f.read())
        if a['encoding'] == 'GB2312':
            a['encoding'] = 'GBK'
    if a['confidence'] < 0.5:
        try:
            logging.warning(f'没能自动识别「{path}」的编码，尝试用默认编码打开。')
            open(path).read()
            return open(path)
        except:
            logging.warning(f'沒能自動識別「{path}」的编码，尝试用utf8编码打开。')
            return open(path, encoding='utf8')
    else:
        return open(path, encoding=a['encoding'])



class vue_ob:
    class vue_sync:
        def __init__(self, master):
            self._master = master
            self._內容 = {}

        def __getattr__(self, x):
            return self._內容[x]

        def __setattr__(self, a, b):
            if a[0] == '_':
                self.__dict__[a] = b
            else:
                self._內容[a] = b
                self._master.更新vue()

    def __init__(self):
        self.vue = self.vue_sync(self)
        self.vue鏈接 = None

    def vue更新(self, 內容):
        self.vue._內容 = 內容

    def 更新vue(self):
        if self.vue鏈接:
            self.vue鏈接.Call(self.vue._內容)

    def vue連接初始化(self, f):
        self.vue鏈接 = f
        self.更新vue()



import math
from math import sin, cos
import functools

import numpy as np


@functools.lru_cache(maxsize=128)
def scale(x, y, z):
    a = np.eye(4, dtype=np.float32)
    a[0, 0] = x
    a[1, 1] = y
    a[2, 2] = z
    return a


@functools.lru_cache(maxsize=128)
def rotate_ax(r, axis: tuple):
    a = np.eye(4, dtype=np.float32)
    a[axis[0], axis[0]] = cos(r)
    a[axis[0], axis[1]] = sin(r)
    a[axis[1], axis[0]] = -sin(r)
    a[axis[1], axis[1]] = cos(r)
    return a


@functools.lru_cache(maxsize=128)
def rotate(angle, v):
    v = np.array(v, dtype=np.float32)
    v /= np.linalg.norm(v)
    a = np.array([0, 0, 1])
    b = np.cross(v, a)
    if np.linalg.norm(b) < np.linalg.norm(v) * np.linalg.norm(a) / 100:
        new_a = np.array([0, 1, 0])
        b = np.cross(a, new_a)
    c = np.cross(v, b)

    rm = np.array([b, c, v])
    arm = np.linalg.inv(rm)

    rm4 = np.eye(4, dtype=np.float32)
    rm4[:3, :3] = rm
    arm4 = np.eye(4, dtype=np.float32)
    arm4[:3, :3] = arm

    return arm4 @ rotate_ax(angle, axis=(0, 1)) @ rm4


@functools.lru_cache(maxsize=128)
def translate(x, y, z):
    a = np.eye(4, dtype=np.float32)
    a[3, 0] = x
    a[3, 1] = y
    a[3, 2] = z
    return a


@functools.lru_cache(maxsize=128)
def perspective(view_distance):
    a = np.eye(4, dtype=np.float32)
    a[2, 2] = 1 / view_distance
    a[3, 2] = -0.0001
    a[2, 3] = 1
    a[3, 3] = 0
    return a



import contextlib
import time


@contextlib.contextmanager
def 计时(名字):
    开始时间 = time.time()
    yield
    print(f'「{名字}」用时:', time.time()-开始时间)


帧率计冷却时间 = 3
帧率计平均用时 = {}
@contextlib.contextmanager
def 帧率计(名字):
    global 帧率计冷却时间, 帧率计平均用时
    开始时间 = time.time()
    yield
    用时 = time.time() - 开始时间
    if 名字 not in 帧率计平均用时:
        帧率计平均用时[名字] = 用时
    帧率计平均用时[名字] = 0.9*帧率计平均用时[名字] + 0.1*用时
    帧率计冷却时间 -= 用时
    if 帧率计冷却时间 < 0:
        帧率计冷却时间 += 1
        for k, v in 帧率计平均用时.items():
            print(f'「{k}」帧率: %.2f' % (1/v))



# -*- coding: utf-8 -*-

from PySide2 import QtWidgets


class DirDialog:
    def __init__(self, window, title):
        self._parentWindow = window
        self._title = title

    def __enter__(self):
        self._dialog = QtWidgets.QFileDialog(parent=self._parentWindow,
                                             caption=self._title)
        self._dialog.setAcceptMode(QtWidgets.QFileDialog.AcceptMode.AcceptOpen)
        self._dialog.setFileMode(QtWidgets.QFileDialog.FileMode.DirectoryOnly)
        self._dialog.setOption(QtWidgets.QFileDialog.ShowDirsOnly, True)
        return self

    def __exit__(self, exc_type, exc, tb):
        if self._dialog.isVisible():
            self._dialog.close()

    def SetPath(self, path):
        self._dialog.setDirectory(path)

    def GetPath(self):
        return self._dialog.directory().absolutePath()

    def ShowModal(self):
        return self._dialog.exec_()


ID_OK = 1



import setuptools


setuptools.setup(
    name='rimo_utils',
    version='1.4.1',
    author='RimoChan',
    author_email='the@librian.net',
    description='RimoChan util.',
    long_description='喵喵喵！',
    long_description_content_type='text/markdown',
    url='https://github.com/RimoChan/rimo_utils',
    packages=['rimo_utils', 'rimo_utils.cef_tools'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        'chardet==3.0.4',
    ],
    python_requires='>=3.6',
)



import logging

from cv2 import *
import numpy as np


def show(img):
    if img.dtype == np.float:
        img = (img * 255).astype(np.uint8)
    cv2.imshow('show', img)
    cv2.waitKey()


def read(img_path):
    img = cv2.imdecode(np.fromfile(img_path, dtype=np.uint8), -1)
    if img is None:
        raise Exception('読み込み失敗')
    return img


def write(img, img_path, param=None):
    with open(img_path, 'wb') as f:
        if img.dtype in (np.float64, np.float32, np.float):
            img = (img * 255).astype(np.uint8)
        if param:
            _, data = cv2.imencode(img_path, img, param)
        else:
            _, data = cv2.imencode(img_path, img)
        f.write(data)


def VideoCapGen(source, size: tuple = None):
    cap = cv2.VideoCapture(source)
    if size:
        x, y = size
        cap.set(cv2.CAP_PROP_FRAME_WIDTH, x)
        cap.set(cv2.CAP_PROP_FRAME_HEIGHT, y)
    while True:
        _, img = cap.read()
        if img is None:
            raise Exception('没图')
        yield img


def circle(img, center, radius, color, thickness=1, lineType=16):
    center = tuple([round(x) for x in center])
    cv2.circle(img, tuple(center), radius, color, thickness, lineType)
    return img


def putText(img, text, org, fontFace, fontScale, color, thickness=1, lineType=16, bottomLeftOrigin=False):
    org = tuple([round(x) for x in org])
    cv2.putText(img, text, org, fontFace, fontScale, color, thickness, lineType, bottomLeftOrigin)
    return img



import logging
import platform
import sys
import os

import wx
from cefpython3 import cefpython as cef


ld_library_path = os.environ.get("LD_LIBRARY_PATH")


def group(url, icon, title, size):
    icon = str(icon)  # pathlib -> str
    sys.excepthook = ExceptHook
    cef.Initialize(settings={}, commandLineSwitches={
        'autoplay-policy': 'no-user-gesture-required',
        'lang': 'zh-CN'
    })
    app = CefApp(url, icon, title, size)
    return app, app.frame.browser


def ExceptHook(exc_type, exc_value, exc_trace):
    '''
    原本的cef.ExceptHook打印不了中文，我拿來魔改一下。
    '''
    import traceback
    if exc_type == SystemExit:
        hint = '[CEF Python] SystemExit.'
        print('\n' + '='*len(hint) + '\n' + hint)
    else:
        hint = '[CEF Python] ExceptHook: catched exception, will shutdown CEF.'
        print('\n' + '='*len(hint) + '\n' + hint)
        msg = ''.join(traceback.format_exception(exc_type, exc_value, exc_trace))
        print('\n' + msg)
    cef.QuitMessageLoop()
    cef.Shutdown()
    os._exit(1)


# Platforms
WINDOWS = (platform.system() == "Windows")
LINUX = (platform.system() == "Linux")
MAC = (platform.system() == "Darwin")

if MAC:
    try:
        # noinspection PyUnresolvedReferences
        from AppKit import NSApp
    except ImportError:
        logging.debug("[wxpython.py] Error: PyObjC package is missing, "
                      "cannot fix Issue #371")
        logging.debug("[wxpython.py] To install PyObjC type: "
                      "pip install -U pyobjc")
        sys.exit(1)

if LINUX:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk, Gdk, GdkX11

# Globals
g_count_windows = 0


def main():
    check_versions()
    sys.excepthook = cef.ExceptHook  # To shutdown all CEF processes on error
    settings = {}
    if MAC:
        # Issue #442 requires enabling message pump on Mac
        # and calling message loop work in a timer both at
        # the same time. This is an incorrect approach
        # and only a temporary fix.
        settings["external_message_pump"] = True
    if WINDOWS:
        # noinspection PyUnresolvedReferences, PyArgumentList
        cef.DpiAware.EnableHighDpiSupport()
    cef.Initialize(settings=settings)
    app = CefApp(False)
    app.MainLoop()
    del app  # Must destroy before calling Shutdown
    if not MAC:
        # On Mac shutdown is called in OnClose
        cef.Shutdown()


def check_versions():
    logging.debug("[wxpython.py] CEF Python {ver}".format(ver=cef.__version__))
    logging.debug("[wxpython.py] Python {ver} {arch}".format(
        ver=platform.python_version(), arch=platform.architecture()[0]))
    logging.debug("[wxpython.py] wxPython {ver}".format(ver=wx.version()))
    # CEF Python version requirement
    assert cef.__version__ >= "66.0", "CEF Python v66.0+ required to run this"


class MainFrame(wx.Frame):

    def __init__(self, url, icon, title, size):
        self.browser = None

        # Must ignore X11 errors like 'BadWindow' and others by
        # installing X11 error handlers. This must be done after
        # wx was intialized.
        if LINUX:
            cef.WindowUtils.InstallX11ErrorHandlers()

        global g_count_windows
        g_count_windows += 1

        if WINDOWS:
            # noinspection PyUnresolvedReferences, PyArgumentList
            logging.debug("[wxpython.py] System DPI settings: %s"
                          % str(cef.DpiAware.GetSystemDpi()))
        if hasattr(wx, "GetDisplayPPI"):
            logging.debug("[wxpython.py] wx.GetDisplayPPI = %s" % wx.GetDisplayPPI())
        logging.debug("[wxpython.py] wx.GetDisplaySize = %s" % wx.GetDisplaySize())

        logging.debug("[wxpython.py] MainFrame DPI scaled size: %s" % str(size))

        wx.Frame.__init__(self, parent=None, id=wx.ID_ANY,
                          title=title)
        # wxPython will set a smaller size when it is bigger
        # than desktop size.
        logging.debug("[wxpython.py] MainFrame actual size: %s" % self.GetSize())

        ic = wx.Icon(icon, wx.BITMAP_TYPE_ICO)
        self.SetIcon(ic)

        self.Bind(wx.EVT_CLOSE, self.OnClose)

        # Set wx.WANTS_CHARS style for the keyboard to work.
        # This style also needs to be set for all parent controls.
        self.browser_panel = wx.Panel(self, size=tuple(size))
        self.browser_panel.Bind(wx.EVT_SIZE, self.OnSize)
        wx.Window.Fit(self)

        if MAC:
            # Make the content view for the window have a layer.
            # This will make all sub-views have layers. This is
            # necessary to ensure correct layer ordering of all
            # child views and their layers. This fixes Window
            # glitchiness during initial loading on Mac (Issue #371).
            NSApp.windows()[0].contentView().setWantsLayer_(True)

        if LINUX:
            self.Show()
            self.embed_browser(url)
        else:
            self.embed_browser(url)
            self.Show()

    def embed_browser(self, url):
        window_info = cef.WindowInfo()
        (width, height) = self.browser_panel.GetClientSize().Get()
        assert self.browser_panel.GetHandle(), "Window handle not available"
        if LINUX:
            handle_to_use = self.browser_panel.GetHandle()
            display = Gdk.Display.get_default()
            window = GdkX11.X11Window.foreign_new_for_display(display, handle_to_use)
            self.gtk_window = gtk_window = Gtk.Window()

            def callback(gtk_window, window):
                print("inside callback")
                gtk_window.set_window(window)
                gtk_window.set_visual(gtk_window.get_screen().lookup_visual(0x21))
            gtk_window.connect("realize", callback, window)
            gtk_window.set_has_window(True)
            gtk_window.show()
            sw = Gtk.ScrolledWindow()
            sw.show()
            gtk_window.add(sw)
            sw.set_visual(sw.get_screen().lookup_visual(0x21))
            self.sw = sw
            self.Show()
            window_info.SetAsChild(sw.get_window().get_xid(), [0, 0, width, height])
        else:
            window_info.SetAsChild(self.browser_panel.GetHandle(), [0, 0, width, height])
        self.browser = cef.CreateBrowserSync(window_info, url=url, browserSettings={'web_security_disabled': True, })
        self.browser.SetClientHandler(FocusHandler())

    def set_browser_object(self, name, obj):
        bindings = cef.JavascriptBindings()
        bindings.SetObject(name, obj)
        self.browser.SetJavascriptBindings(bindings)

    def OnSetFocus(self, _):
        if not self.browser:
            return
        if WINDOWS:
            cef.WindowUtils.OnSetFocus(self.browser_panel.GetHandle(),
                                       0, 0, 0)
        self.browser.SetFocus(True)

    def OnSize(self, _):
        if not self.browser:
            return
        if WINDOWS:
            cef.WindowUtils.OnSize(self.browser_panel.GetHandle(),
                                   0, 0, 0)
        elif LINUX:
            (x, y) = (0, 0)
            (width, height) = self.browser_panel.GetSize().Get()
            self.browser.SetBounds(x, y, width, height)
            self.sw.get_window().move_resize(x, y, width, height)
        self.browser.NotifyMoveOrResizeStarted()

    def OnClose(self, event):
        logging.debug("[wxpython.py] OnClose called")
        if not self.browser:
            # May already be closing, may be called multiple times on Mac
            return

        if MAC:
            # On Mac things work differently, other steps are required
            self.browser.CloseBrowser()
            self.clear_browser_references()
            self.Destroy()
            global g_count_windows
            g_count_windows -= 1
            if g_count_windows == 0:
                cef.Shutdown()
                wx.GetApp().ExitMainLoop()
                # Call _exit otherwise app exits with code 255 (Issue #162).
                # noinspection PyProtectedMember
                os._exit(0)
        else:
            # Calling browser.CloseBrowser() and/or self.Destroy()
            # in OnClose may cause app crash on some paltforms in
            # some use cases, details in Issue #107.
            self.browser.ParentWindowWillClose()
            event.Skip()
            self.clear_browser_references()

    def clear_browser_references(self):
        # Clear browser references that you keep anywhere in your
        # code. All references must be cleared for CEF to shutdown cleanly.
        self.browser = None

    def toggleFullScreen(self):
        if self.IsFullScreen():
            self.ShowFullScreen(False)
        else:
            self.ShowFullScreen(True)


class FocusHandler(object):
    def OnGotFocus(self, browser, **_):
        # Temporary fix for focus issues on Linux (Issue #284).
        if LINUX:
            logging.debug("[wxpython.py] FocusHandler.OnGotFocus:"
                          " keyboard focus fix (Issue #284)")
            browser.SetFocus(True)


class CefApp(wx.App):
    def __init__(self, url, icon, title, size):
        self.url, self.icon, self.title, self.size = url, icon, title, size

        self.timer = None
        self.timer_id = 1
        self.is_initialized = False

        super(CefApp, self).__init__(redirect=False)

    def OnPreInit(self):
        super(CefApp, self).OnPreInit()
        # On Mac with wxPython 4.0 the OnInit() event never gets
        # called. Doing wx window creation in OnPreInit() seems to
        # resolve the problem (Issue #350).
        if MAC and wx.version().startswith("4."):
            logging.debug("[wxpython.py] OnPreInit: initialize here"
                          " (wxPython 4.0 fix)")
            self.initialize()

    def OnInit(self):
        self.initialize()
        return True

    def initialize(self):
        if self.is_initialized:
            return
        self.is_initialized = True
        self.create_timer()
        self.frame = MainFrame(self.url, self.icon, self.title, self.size)
        self.SetTopWindow(self.frame)
        self.frame.Show()

    def create_timer(self):
        # See also "Making a render loop":
        # http://wiki.wxwidgets.org/Making_a_render_loop
        # Another way would be to use EVT_IDLE in MainFrame.
        self.timer = wx.Timer(self, self.timer_id)
        self.Bind(wx.EVT_TIMER, self.on_timer, self.timer)
        self.timer.Start(10)  # 10ms timer

    def on_timer(self, _):
        cef.MessageLoopWork()

    def OnExit(self):
        self.timer.Stop()
        return 0


if __name__ == '__main__':
    main()
